#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

// Define the number of threads per block; must be a multiple of 32.
#define THREADS_PER_BLOCK 128

// This kernel replaces any shared memory reductions with warp-level __shfl_down_sync primitives.
// Each block handles one batch row, and each warp within the block computes one output neuron's dot product.

template <typename scalar_t>
__global__ void mlp_forward_kernel_warp_shfl(
    const scalar_t* __restrict__ input,
    const scalar_t* __restrict__ weight,
    const scalar_t* __restrict__ bias,
    scalar_t* __restrict__ output,
    int batch_size,
    int in_features,
    int out_features) {

    // Each block processes one row of the batch
    int batch_row = blockIdx.x;
    
    // Each warp (of 32 threads) calculates one output neuron.
    int warp_id = threadIdx.x / 32;  // warp index within the block
    int lane = threadIdx.x % 32;       // lane index within the warp

    // Determine the neuron index computed by this warp across the output
    int neurons_per_block = blockDim.x / 32;
    int neuron = blockIdx.y * neurons_per_block + warp_id;

    if (batch_row >= batch_size || neuron >= out_features) return;

    scalar_t sum = 0;
    // Each thread processes elements in a strided loop over in_features
    for (int i = lane; i < in_features; i += 32) {
        sum += input[batch_row * in_features + i] * weight[neuron * in_features + i];
    }

    // Use warp-level reduction to sum across lanes (no shared memory required)
    for (int offset = 16; offset > 0; offset /= 2) {
        sum += __shfl_down_sync(0xffffffff, sum, offset);
    }

    // The first lane writes the result, adding the bias
    if (lane == 0) {
        output[batch_row * out_features + neuron] = sum + bias[neuron];
    }
}

// Optimized ReLU activation kernel using a stride loop to cover all elements
template <typename scalar_t>
__global__ void relu_kernel_shfl(
    scalar_t* __restrict__ data,
    int size) {
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int stride = blockDim.x * gridDim.x;
    for (int i = idx; i < size; i += stride) {
        data[i] = data[i] > 0 ? data[i] : static_cast<scalar_t>(0);
    }
}

// Main MLP forward pass function that applies the kernel layer by layer.
// This implementation uses warp-level primitives exclusively for reduction, eliminating shared memory usage.

torch::Tensor mlp_cuda_forward(
    torch::Tensor input,
    std::vector<torch::Tensor> weights,
    std::vector<torch::Tensor> biases) {

    auto device = input.device();
    int num_layers = weights.size();
    torch::Tensor current = input;

    for (int layer = 0; layer < num_layers; layer++) {
        int batch_size = current.size(0);
        int in_features = current.size(1);
        int out_features = weights[layer].size(0);

        auto output = torch::empty({batch_size, out_features},
                                   torch::dtype(current.dtype()).device(device));

        // Configure the grid: one block per batch row and groups of warps for output neurons.
        dim3 block(THREADS_PER_BLOCK);
        int neurons_per_block = THREADS_PER_BLOCK / 32; // one warp per neuron
        dim3 grid(batch_size, (out_features + neurons_per_block - 1) / neurons_per_block);

        AT_DISPATCH_FLOATING_TYPES(current.scalar_type(), "mlp_forward_kernel_warp_shfl", ([&] {
            mlp_forward_kernel_warp_shfl<scalar_t><<<grid, block>>>(
                current.data_ptr<scalar_t>(),
                weights[layer].data_ptr<scalar_t>(),
                biases[layer].data_ptr<scalar_t>(),
                output.data_ptr<scalar_t>(),
                batch_size,
                in_features,
                out_features
            );
        }));

        // Apply ReLU activation for non-final layers
        if (layer < num_layers - 1) {
            int size = batch_size * out_features;
            int threads_per_block_relu = 256;
            int num_blocks_relu = (size + threads_per_block_relu - 1) / threads_per_block_relu;
            AT_DISPATCH_FLOATING_TYPES(output.scalar_type(), "relu_kernel_shfl", ([&] {
                relu_kernel_shfl<scalar_t><<<num_blocks_relu, threads_per_block_relu>>>(
                    output.data_ptr<scalar_t>(),
                    size
                );
            }));
        }
        
        current = output;
    }

    return current;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &mlp_cuda_forward, "MLP forward (CUDA warp shfl optimized)");
}
